
//	Copyright H. Giesen, University of Koblenz-Landau 1995

#import "Shelf.h"
#import "Controller.h"
#import "DragView.h"
#import "BibliographicFile.h"
#import "Preferences.h"


#define DRAGVIEWTAG	7
#define W_SPACE		10.0
#define H_SPACE		10.0
#define L_MARGIN	5.0


@implementation Shelf

- (void) _prRect:(const NXRect *)r
{
	fprintf( stderr, "x=%d y=%d w=%d h=%d\n",
		(int)r->X, (int)r->Y, (int)r->W, (int)r->H );
}


- initFrame:(const NXRect *)frameRect
/*
%initFrame% initializes the view and registers the Pasteboard types that the shelf will accept in an image-dragging session. The shelf accepts only the NXFilenamePboardType.
Returns self.
*/
{
	[self registerForDraggedTypes:&NXFilenamePboardType count:1];;
	MyListPboardType = NXUniqueString( "My List Pboard Type" );
	[super initFrame:frameRect];
	[self setFlipped:YES];
	preferences = [Preferences new];
	return  self;
}


- (void) setBoxPrototype:(NXStream *)aStream
/* New boxes for the shelf are generated from %aStream% by calling
the function NXReadObjectFromBuffer() using the buffer of %aStream%.
See also readBoxFromStream.
*/
{
	protypeStream = aStream;
}


- (BOOL)loadShelfFromFile:(const char *)file
{
	int		newInx;
	int		okFlag;
	Box		*newBox;
	char	buffer[1024];
	char 	*first, *last, *comment;
	FILE	*fp;
	
	fp = fopen( file, "r" );
	
	if( fp==NULL ) return NO;
	if( protypeStream==NULL ) return NO;

	while( fgets( buffer, 512, fp ) ){
		first = strchr( buffer, '"');	// first occurence of '"'
		if( first==NULL ) continue;
		last = strrchr( buffer, '"');	// last occurence of '"'
		if( last==NULL ) continue;
		*last = '\0';
		// commentline ?
		comment = strchr( buffer, '%');	// first occurence of '%'
		if( comment && (comment<first) ) continue;
		// filename is ok
		newBox = [self readBoxFromStream:protypeStream];

		newInx = [self insertBox:newBox];
		[[newBox findViewWithTag:DRAGVIEWTAG] setFilename:first+1];
		last++;		//first character behind '\0', was '"'
		while( *last ){
			int	rtn;
			if( *last=='+' ){	// open the file
				shelfBox[newInx].state = 1;
				rtn = [NXApp openFile:first+1 ok:&okFlag];
				break;
			}
			if( *last=='-' ) break;
			last++;
		}
	}
	fclose( fp );
	return YES;
}


/***********	find box or empty slot	************/

- (int)indexOfBox:(Box *)box
{
	int	j;

	for( j=0; j<count; j++ ){
		if( shelfBox[j].theBox == box ) return j;
	}
	return -1;
}


- (int)indexOfBoxWithIcon:(DragView *)icon
{
	int	j;

	for( j=0; j<count; j++ ){
		if( shelfBox[j].theView == icon ) return j;
	}
	return -1;
}


- (int)indexOfBoxForFile:(BibliographicFile *)aFile
{
	int	j;

	for( j=0; j<count; j++ ){
		if( shelfBox[j].theFile == aFile ) return j;
	}
	return -1;
}


- (int)indexOfBoxNamed:(const char *)fileName
{
	int	j;

	for( j=0; j<count; j++ ){
		if( [shelfBox[j].theView fileName]==NULL ) continue;
		if( strcmp([shelfBox[j].theView fileName], fileName)==0 ) return j;
	}
	return -1;
}


- (int)indexOfEmptySlot
{
	int	j;

	for( j=0; j<count; j++ ){
		if( shelfBox[j].theBox==nil ) return j;
	}
	return -1;
}


/*****	 properties of an icon		****/

- (const char *)filenameOfIconAt:(int)inx
{
	if( (inx<0) || (inx>=count) ) return NULL;
	if( shelfBox[inx].theView==nil ) return NULL;
	return [shelfBox[inx].theView fileName];
}


- (int)getstateOfIconAt:(int)inx
{
	if( (inx<0) || (inx>=count) ) return -1;
	return shelfBox[inx].state;
}


- setState:(int)state forIconAt:(int)inx
{
	if( (inx<0) || (inx>=count) ) return nil;
	//fprintf( stderr, "shell new state %d at %d\n", state, inx );
	shelfBox[inx].state = state;
	return self;
}



/********		notification methods	*******/

static const char * dragTypes[3];

- attachFile:(const char *)fName toObject:(BibliographicFile *)obj
{
	int j = [self indexOfBoxNamed:fName];
	if( j<0 ) return nil;
	
	[shelfBox[j].theView setFileObject:obj];
	dragTypes[0] = MyListPboardType;
	dragTypes[1] = NXAsciiPboardType;
	[shelfBox[j].theView registerForDraggedTypes:dragTypes count:2];
	shelfBox[j].theFile = obj;		
	[shelfBox[j].theView display];
	return self;
}


- attachObject:(BibliographicFile *)obj toFile:(const char *)fName
{
	int j = [self indexOfBoxForFile:obj];
	if( j<0 ) return nil;
		
	[shelfBox[j].theView setFileObject:obj];		
	[shelfBox[j].theView display];

	[preferences shelfDidChange];
	return self;
}


- detachObject:(BibliographicFile *)obj
{
	int j = [self indexOfBoxForFile:obj];
	if( j<0 ) return nil;
		
	[shelfBox[j].theView unregister];
	shelfBox[j].theFile = nil;

	return self;
}



/***********	remove box identified by ...	************/


- (void) removeBoxAt:(int)inx
{
	if( inx<0 ) return;

	[shelfBox[inx].theView clear];
	[shelfBox[inx].theView unregisterDraggedTypes];

	[shelfBox[inx].theBox removeFromSuperview];
	[NXApp delayedFree:shelfBox[inx].theBox];
	shelfBox[inx].theBox = nil;
	shelfBox[inx].theFile = nil;
	shelfBox[inx].theView = nil;

	if( inx==count-1 ){	// compact if possible
		while( count && (shelfBox[count-1].theBox==nil) ) count--;
		[self redrawShelf];
	}
	else [self display];
}


- (void) removeBoxWithIcon:(DragView *)icon
{
	[self removeBoxAt:[self indexOfBoxWithIcon:icon]];
	[self display];
	[preferences shelfDidChange];
}


- (Box *) readBoxFromStream:(NXStream *)theStream
{
	char	*buffer;
	int		len, maxlen;
	Box		*box;

	NXGetMemoryBuffer(theStream, &buffer, &len, &maxlen);
	box = NXReadObjectFromBuffer( buffer, len);

	[box getFrame:&boxRect];
	[box setBorderType:NX_NOBORDER];
	
	return box;
	
}


- (void) sizeShelf
{
	NXRect	r;

	if( scrollView==nil ) 
		scrollView = [[self superview] superview];
	[[self superview] getFrame:&r];
	iconsPerRow = (int)((r.W - L_MARGIN + W_SPACE)/(boxRect.W + W_SPACE));
	[self sizeTo:iconsPerRow*(boxRect.W + W_SPACE)
			 	:(((count-1)/iconsPerRow)+1)*(boxRect.H + H_SPACE)];
	[scrollView setLineScroll:(boxRect.H + H_SPACE)];
}


- (void) setCoordinates:(NXRect *)r at:(int)inx
{
	r->X = (inx%iconsPerRow)*(r->W + W_SPACE) + L_MARGIN;
	r->Y = (inx/iconsPerRow)*(r->H + H_SPACE);
}


- (void) redrawShelf
{
	Box		*box;
	int		in, out;

	[self sizeShelf];
	for( in=0, out=0; in<count; in++){
		box =  shelfBox[in].theBox;
		if( box==nil ) continue;
		[self setCoordinates:&boxRect at:out];
		[box setFrame:&boxRect];
		shelfBox[out] = shelfBox[in];
		out++;
	}
	if( count==out ) return;
	count = out;
	[self sizeShelf];
	[preferences shelfDidChange];
}


- (int)insertBox:(Box *)box at:(int)inx
{
	if( (inx<0) || (inx>=MAXCOUNT) ){
		fprintf( stderr, "Index %d ungueltig\n", inx );
		return -1;
	}

	[self setCoordinates:&boxRect at:inx];
	[box setFrame:&boxRect];

	shelfBox[inx].theView = [box findViewWithTag:DRAGVIEWTAG];
	shelfBox[inx].theBox = box;
	shelfBox[inx].theFile = nil;
	shelfBox[inx].state = 0;
	[shelfBox[inx].theView setShelf:self];

	[self addSubview:box];
	[self scrollRectToVisible:&boxRect];
	[preferences shelfDidChange];
	return inx;
}


- (int)insertBox:(Box *)box
{
	int	inx = [self indexOfEmptySlot];

	if( inx>=0 ){
		[self insertBox:box at:inx];
	}
	else {
		if( count>=MAXCOUNT ){
			fprintf( stderr, "zuviele icons\n" );
			return -1;
		}
		count++;
		[self sizeShelf];
		inx = [self insertBox:box at:count-1];	// last entry in table
	}
	[self display];

	return inx;
}


- (int)count
{
	return count;
}


- windowDidResize:sender
{
	[self redrawShelf];
	return self;
}



/************		dragging		************/

// includesType() returns YES if type is included within types.

static BOOL includesType (const NXAtom *types, NXAtom type)
{
    if (types) while (*types){
		if (*types++ == type) return YES;
	}
    return NO;
}


- (NXDragOperation)draggingEntered:(id <NXDraggingInfo>)sender
{
    return [self draggingUpdated:sender];	    
}


- (NXDragOperation)draggingUpdated:(id <NXDraggingInfo>)sender
{
	Pasteboard *pboard = [sender draggingPasteboard];
	char *data, *lastDot, *file, *tab;
	int length;

	if (includesType([pboard types], MyListPboardType)) {
		return NX_DragOperationNone;
	}

    if( ([sender draggingSourceOperationMask] & NX_DragOperationGeneric) 
	  ||([sender draggingSourceOperationMask] & NX_DragOperationCopy) ){
	  
		if( includesType([pboard types], NXFilenamePboardType) ){
    		[pboard readType:NXFilenamePboardType data:&data length:&length];
			// is it a list of files ?
			file = data;
			while (file) {
				if( tab = strchr(file, '\t') ) { *tab = '\0'; }
				lastDot = rindex( file, '.' );
				if( lastDot ){ 
					if( strcmp( lastDot, ".bib" ) ) {	// only .bib allowed
						return NX_DragOperationNone;
					}
				}
				else return NX_DragOperationNone;	// no suffix -> invalid
				// is the file already on the shelf ?
				if( [self indexOfBoxNamed:file]>=0 ){	// yes
					return NX_DragOperationNone;
				}
				file = (tab) ? ++tab : NULL;
			}
			return NX_DragOperationGeneric;
		}
    }	
    return NX_DragOperationNone;	    
}


- (BOOL)performDragOperation:(id <NXDraggingInfo>)sender
{
	char *data, *file, *tab;
	int length;

	Pasteboard *pboard = [sender draggingPasteboard];

	if( includesType([pboard types], NXFilenamePboardType) ){
		[pboard readType:NXFilenamePboardType data:&data length:&length];

		file = data;
		while (file) {
			Box					*newBox;
			DragView			*newView;
			BibliographicFile	*fileObj;

			if (tab = strchr(file, '\t')) {
				*tab = '\0';
			}
			newBox = [self readBoxFromStream:protypeStream];
			newView = [newBox findViewWithTag:DRAGVIEWTAG];
			[newView setFilename:file];
			[self insertBox:newBox];
			// ask if this file is open in this application
			fileObj = [[NXApp delegate] objectForFileNamed:file];
			if( fileObj )[self attachFile:file toObject:fileObj];
			file = (tab) ? ++tab : NULL;
		}
		return YES;
	}

	return NO;
}


- draggingExited:sender
{
    return self;
}


@end
